import { AddCdnDomainRequest } from 'tencentcloud-sdk-nodejs/tencentcloud/services/cdn/v20180606/cdn_models';
import { pascalCaseProperty } from '@tencent-serverless-framework/common';
import { CdnDeployInputs, CertInputs, OriginInputs } from './interface';
import * as fs from 'fs';
import * as path from 'path';

/** 用于获取证书文件信息 */
function getCertPathContent(target: string) {
  let content = '';

  try {
    const stat = fs.statSync(target);
    if (stat.isFile()) {
      if (path.isAbsolute(target)) {
        content = fs.readFileSync(target, 'base64');
      } else {
        content = fs.readFileSync(path.join(process.cwd(), target), 'base64');
      }
    }
  } catch (e) {
    // target is string just return
    content = target;
  }
  return content;
}

/**
 * 格式化证书
 * @param {object} certInfo cert info
 */
function certInputs(certInfo: CertInputs): CertInputs {
  if (certInfo.certId) {
    return {
      certId: certInfo.certId,
    };
  }
  return {
    certificate: getCertPathContent(certInfo.certificate),
    privateKey: getCertPathContent(certInfo.privateKey),
    message: certInfo.remarks ?? '',
  };
}

/** 格式化 CDN 源信息 */
export function originInputs(origin: OriginInputs): OriginInputs {
  const originInfo = {
    origins: origin.origins,
    originType: origin.originType,
    originPullProtocol: origin.originPullProtocol,
    serverName: origin.serverName,
  };
  if (origin.originType === 'cos') {
    return {
      ...originInfo,
      serverName: origin.origins[0],
      cosPrivateAccess: 'off',
    };
  }
  if (origin.originType === 'domain') {
    if (origin.backupOrigins) {
      return {
        ...originInfo,
        backupOrigins: origin.backupOrigins,
        backupOriginType: origin.backupOriginType,
        backupServerName: origin.backupServerName,
      };
    }
  }
  return originInfo;
}

/** DeployInputs 转 AddCdnDomainRequest */
function deployInputs(inputs: CdnDeployInputs): AddCdnDomainRequest {
  if (inputs.https) {
    const hs = inputs.https;
    inputs.https = {
      switch: hs.switch ?? 'on',
      http2: hs.http2 || 'off',
      ocspStapling: hs.ocspStapling || 'off',
      verifyClient: hs.verifyClient || 'off',
      certInfo: certInputs(hs.certInfo),
    };
  }
  if (inputs.forceRedirect && inputs.https) {
    inputs.forceRedirect = {
      switch: inputs.forceRedirect.switch ?? 'on',
      redirectStatusCode: inputs.forceRedirect.redirectStatusCode ?? 301,
      redirectType: 'https',
    };
  }
  return pascalCaseProperty(inputs);
}

export const formatter = {
  originInputs,
  certInputs,
  deployInputs,
};
